// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "SceneItem.hpp"
#include "event/Event.hpp"

namespace Dawn {
  class DawnGame;

  class Scene {
    private:
      sceneitemid_t nextId;
      std::map<sceneitemid_t, std::shared_ptr<SceneItem>> items;
      std::map<sceneitemid_t, std::shared_ptr<SceneItem>> itemsNotInitialized;

    public:
      DawnGame &game;
      Event<> eventSceneUpdate;
      Event<> eventSceneUnpausedUpdate;

      /**
       * Construct a new Scene instance.
       * 
       * @param game Reference to the game that this scene belongs to.
       */
      Scene(DawnGame &game);

      /**
       * Perform a one frame synchronous tick on the current scene. This may
       * change in future to be more event-like.
       */
      void update();

      /**
       * Create a Scene Item object and add to the current scene.
       * 
       * @return A shared pointer to the created SceneItem.
       */
      std::shared_ptr<SceneItem> createSceneItem();

      /**
       * Finds an existing component on the scene (Root Level Only) that has a
       * component matching the given component type. Returns nullptr if no item
       * with the specified component could be found.
       * 
       * @tparam Component type to look for.
       * @return Pointer to the found component (and by extension the item).
       */
      template<class T>
      std::shared_ptr<T> findComponent() {
        auto it = this->items.begin();
        while(it != this->items.end()) {
          auto component = it->second->getComponent<T>();
          if(component != nullptr) return component;
          ++it;
        }
        return nullptr;
      }

      /**
       * Finds all exisitng components on the scene (Root Level Only) that has a
       * matching component of the given component type.
       * 
       * @tparam Component type to look for.
       * @return List of matching compnoents.
       */
      template<class T>
      std::vector<std::shared_ptr<T>> findComponents() {
        std::vector<std::shared_ptr<T>> components;
        auto it = this->items.begin();
        while(it != this->items.end()) {
          auto component = it->second->getComponent<T>();
          if(component != nullptr) components.push_back(component);
          ++it;
        }
        return components;
      }

      /**
       * Destroys a previously initialized Scene.
       */
      ~Scene();
  };
}